package GUI::MainScreen;

use strict;
use vars qw($VERSION);
$VERSION = '0.01';

use GUI::Help;
use GUI::TaskEditor;
use GUI::ResourceEditor;
use GUI::ResourceUsageEditor;

use RMA::System;
use RMA::Task;
use RMA::Resource;
use RMA::Analyzer;

use Class::MethodMaker
  new_with_init => 'new',
  new_hash_init => 'hash_init',
  get_set => [ 'title', 'system', 'state', 'background', 'foreground', 'font' ],
  hash => [ 'interface' ],
;

sub init {
  my $self = shift;
  $self->hash_init(@_);

  $self->load_all;

  # initialize the window
  $self->init_window;

  # makes it easier to test - remove later
  # $self->load_system;

  return $self;
}

sub init_window {
  my $self = shift;

  my $interface = $self->interface;
  if (!exists($interface->{top})) {
    $interface->{top} = MainWindow->new();
  }
  # get rid of any existing contents if passed a toplevel
  my $top = $interface->{top};
  foreach my $child ($top->children()) {
    eval {
      $top->destroy($child);
    }
  }

  # setup the title
  if (defined($self->title)) {
    $top->title($self->title);
  } else {
    $top->title('gRMA');
  }

  # just a hint - wm's with a better clue than mine will ignore it anyway.
  $top->geometry('+70+70');

  my $menu_frame = $top->Frame()->grid(-row => 0, -column => 0, -sticky => 'new' );
  $interface->{menu_frame} = $menu_frame;
  my $client_frame = $top->Frame(-border => 1, -relief => 'sunken', 
                                 -width => 700, -height => 400)
    ->grid(-row => 1, -column => 0, -sticky => 'nsew' );
  $interface->{client_frame} = $client_frame;
  my $status_frame = $top->Frame(-border => 3, -relief => 'sunken')
    ->grid(-row => 2, -column => 0, -sticky => 'sew' );
  $interface->{status} = $status_frame->Label(-text => 'Welcome to gRMA')->pack(-side => 'left', -fill => 'x');

  $top->gridColumnconfigure(0, -weight => 1);
  $top->gridRowconfigure(0, -weight => 0);
  $top->gridRowconfigure(1, -weight => 1);
  $top->gridRowconfigure(2, -weight => 0);

  my $file_menubutton = $menu_frame->Menubutton(-text => 'File', -relief => 'flat')
    ->pack(-side => 'left');
  $file_menubutton->command(-label => 'New System', 
                            -command => sub {
                              $self->new_system; 
                            } 
                           );
  $file_menubutton->command(-label => 'Load System', 
                            -command => sub {
                              $self->load_system
                            } 
                           );
  $file_menubutton->command(-label => 'Save System', 
                            -command => sub {
                              $self->save_system
                            } 
                           );
  $file_menubutton->command(-label => 'Remove System', 
                            -command => sub {
                              $self->remove_system
                            } 
                           );
  $file_menubutton->separator();
  $file_menubutton->command(-label => 'Exit', 
                            -command => [\&exit_program, $self],
                           );

  my $help_menubutton = $menu_frame->Menubutton(-text => 'Help', -relief => 'flat')
    ->pack(-side => 'right');
  $help_menubutton->command(-label => 'Usage Guide', 
                            -command => [\&showHelp, 'usage']
                           );
  $help_menubutton->command(-label => 'Task Editor', 
                            -command => [\&showHelp, 'TaskEditor']
                           );
  $help_menubutton->command(-label => 'Resource Editor', 
                            -command => [\&showHelp, 'ResourceEditor']
                           );
  $help_menubutton->command(-label => 'Resource Usage Layout', 
                            -command => [\&showHelp, 'ResourceUsageEditor']
                           );
  $help_menubutton->command(-label => 'Timing Analysis', 
                            -command => [\&showHelp, 'TimingAnalysis']
                           );
  $help_menubutton->separator();
  $help_menubutton->command(-label => 'Show License', 
                            -command => [\&showHelp, 'license']
                           );
  $help_menubutton->command(-label => 'About...', 
                            -command => [\&showHelp, 'about']
                           );


  my $view_menubutton = $menu_frame->Menubutton(-text => 'View', -relief => 'flat')
    ->pack(-side => 'left');
  $view_menubutton->radiobutton(-label => 'Task List', 
                                -command => [ \&change_view_state, $self ],
                                -variable => \$interface->{view_state},
                                -value => 'task list' );
  $view_menubutton->radiobutton(-label => 'Resource List', 
                                -command => [ \&change_view_state, $self ],
                                -variable => \$interface->{view_state},
                                -value => 'resource list' );
  $view_menubutton->radiobutton(-label => 'Resource Usage Layout', 
                                -command => [ \&change_view_state, $self ],
                                -variable => \$interface->{view_state},
                                -value => 'resource usage layout' );
  $view_menubutton->radiobutton(-label => 'Timing Analysis', 
                                -command => [ \&change_view_state, $self ],
                                -variable => \$interface->{view_state},
                                -value => 'timing analysis' );
  #$view_menubutton->radiobutton(-label => 'Execution Graph', 
  #                              -command => [ \&change_view_state, $self ],
  #                              -variable => \$interface->{view_state},
  #                              -value => 'execution graph' );

  $top->update;

  return $self;
}

###
# gui event handlers
###

# gets called whenever the gui switches to a new system
sub switch_systems {
  my $self = shift;
  $self->interface('top')->configure(-title => 'gRMA : ' . $self->system->name);
  defined($self->interface('view_state')) || ($self->interface('view_state', 'task list'));
  $self->change_view_state;
  $self->interface('top')->update;
}


sub change_view_state {
  my $self = shift;
  my $view_state = $self->interface('view_state');

  if ($view_state eq 'task list') {
    $self->show_task_list($self->interface('client_frame'));
  } elsif ($view_state eq 'resource list') {
    $self->show_resource_list($self->interface('client_frame'));
  } elsif ($view_state eq 'resource usage layout') {
    $self->show_resource_usage_layout($self->interface('client_frame'));
  } elsif ($view_state eq 'timing analysis') {
    $self->show_timing_analysis($self->interface('client_frame'));
  #} elsif ($view_state eq 'execution graph') {
  #  
  } else {
    die "Unknown view_state $view_state.";
  }

}

# called to emit status messages
sub status {
  my $self = shift;
  my $msg = shift;
  $self->interface('status')->configure(-text => $msg);
}

###
# data loading and storing functions
###

# load_all is called by init at load time
sub load_all {
  opendir(DATA, $::DATA) || die("Unable to open $::DATA : $!");
  foreach my $system_id (readdir(DATA)) {
    open(SYSTEMDB, $::DATA . '/' . $system_id . '/system.db') || next;
    open(TASKSDB, $::DATA . '/' . $system_id . '/tasks.db') || next;
    open(RESOURCESDB, $::DATA . '/' . $system_id . '/resources.db') || next;
    
    my ($buffer, $dump);
    while(read(SYSTEMDB, $buffer, 1024)) {$dump .= $buffer;}
    my @dumps = split("### BREAK ###\n", $dump);
    foreach my $dump (@dumps) {
      if (defined($dump) && length($dump)) { 
        RMA::System->load($dump);
      }
    }
    close(SYSTEMDB);
    $dump = "";

    while(read(TASKSDB, $buffer, 1024)) {$dump .= $buffer;}
    @dumps = split("### BREAK ###\n", $dump);
    foreach my $dump (@dumps) {
      if (defined($dump) && length($dump)) { 
        RMA::Task->load($dump);
      }
    }
    close(TASKSDB);
    $dump = "";

    while(read(RESOURCESDB, $buffer, 1024)) {$dump .= $buffer;}
    @dumps = split("### BREAK ###\n", $dump);
    foreach my $dump (@dumps) {
      if (defined($dump) && length($dump)) { 
        RMA::Resource->load($dump);
      }
    }
    close(RESOURCESDB);
  }
  closedir(DATA);
}

# saves a system to disk - called by File menu and dialog boxes galore
sub save_system {
  my $self = shift;

  my $system = $self->system;
  if (!defined($system)) {
    $self->interface('top')->messageBox(-message => 'Cannot save System!  No system is loaded.',
                           -title => 'Problem Saving System',
                           -type => 'OK');
    return;
  } 
                           
  # make directory if it doesn't already exist
  mkdir $::DATA . '/' . $system->id, 0700;
  
  # save the tasks
  open (DATA, '>' . $::DATA . '/' . $system->id . '/tasks.db') ||
    die("Could not open tasks.db : $!");
  if (defined($system->tasks)) {
    foreach my $task_id ($system->tasks_keys) {
      my $task = RMA::Task->find($task_id);
      print DATA $task->dump;
      print DATA "### BREAK ###\n";
    }
  }

  close DATA;

  # save the resources
  open (DATA, '>' . $::DATA . '/' . $system->id . '/resources.db') ||
    die("Could not open resources.db : $!");
  if (defined($system->resources)) {
    foreach my $resource_id ($system->resources_keys) {
      my $resource = RMA::Resource->find($resource_id);
      print DATA $resource->dump;
      print DATA "### BREAK ###\n";
    }
  }
  close DATA;

  # save the system
  open (DATA, '>' . $::DATA . '/' . $system->id . '/system.db') ||
    die("Could not open system.db : $!");
  print DATA $system->dump;
  close DATA;

  $self->status('Save successful.');
}

# loads a system from disk - called by File menu
sub load_system {
  my $self = shift;

  my $system = $self->system;

  # save the current system?
  if (defined($system)) {
    my $answer = $self->interface('top')->messageBox(
                                                     -message => 
                                                     'Save Current System?',
                                                     -title => 'Save?',
                                                     -type => 'YesNo'
                                                    );
    if ($answer =~ /yes/i) {
      $self->save_system;
    }    
  } 

  # show a list of available systems
  
  # show the list
  
  # open box to select system to load
  my $top = $self->interface('top')->Toplevel(-title => 'Load System');
  $top->withdraw;
  
  # should change to be center of client area...
  $top->geometry('+100+100');

  $top->Label(-text => 'Select a system to load:')
    ->grid(-row => 0, -column => 0, -sticky => 'nsew');
  my $listbox = $top->Scrolled('Listbox', -width => 20, -height => 10, -scrollbars => 'se')
    ->grid(-row => 1, -column => 0, -sticky => 'nesw');

  # get a sorted list of RMA::Systems 
  my @sorted_systems = sort {lc($a->name) cmp lc($b->name);} (values %{RMA::System->find()});
  foreach my $system (@sorted_systems) {
    $listbox->insert('end', $system->name);
  }

  $listbox->bind('<Double-Button-1>', sub {
                   my $which = $listbox->curselection;
                   if (defined($which)) {
                     $self->system($sorted_systems[$which]);
                     $top->grabRelease;
                     $top->destroy;
                     $self->switch_systems;
                     $self->status('Load successful.');
                   }
                 }
                );

  $top->Button(-text => 'CANCEL', 
               -command => 
               sub {
                   $top->grabRelease;
                   $top->destroy;
                   $self->interface('top')->update;
                 }               
              )
    ->grid(-row => 2, -column => 0, -sticky => 'new');

  $top->deiconify; 
  $top->grab;
}

# removes the current system from memory and disk
sub remove_system {
  my $self = shift;
  my $system = $self->system;

  # really?
  my $answer = $self->interface('top')->messageBox(
                                                   -message => 
                                                   'Really Remove Current System?',
                                                   -title => 'Remove System?',
                                                   -type => 'YesNo'
                                                    );
  if ($answer =~ /no/i) {
    return;
  }
  
  # remove system from RMA::System find hash - essentially unlink it
  # for the garbage collector to (hopefully!) harvest.
  my $find_hash = RMA::System->find;
  delete($find_hash->{$system->id});

  # delete the files associated with the system.
  unlink($::DATA . '/' . $system->id . '/resources.db');
  unlink($::DATA . '/' . $system->id . '/tasks.db');
  unlink($::DATA . '/' . $system->id . '/system.db');
  rmdir($::DATA . '/' . $system->id);

  # finally, set current system to undef
  $self->clear_system;
  $self->status('System removed successfully.');
}

# creates a new System for the GUI to work on
sub new_system {
  my $self = shift;
  my $interface = $self->interface;

  # check to save existing
  if (defined($self->system)) {
    my $answer = $self->interface('top')->messageBox(
                                                     -message => 
                                                     'Save Current System?',
                                                     -title => 'Save?',
                                                     -type => 'YesNo'
                                                    );
    if ($answer =~ /yes/i) {
      $self->save_system;
    }
  }
  
  $self->system( RMA::System->new);

  # open box to enter the new system name
  my $top = $self->interface('top')->Toplevel(-title => 'New System');
  $top->withdraw;
  
  # should change to be center of client area...
  $top->geometry('+100+100');

  $top->Label(-text => 'Enter a name for the new system: ')
    ->grid(-row => 0, -column => 0, -sticky => 'nsew');
  $top->Entry(-textvariable => \$interface->{new_system_name},
              -width => 20)
    ->grid(-row => 0, -column => 1, -sticky => 'new');
  $top->Button(-text => 'OK', 
               -command => 
               sub {
                 if (defined($interface->{new_system_name}) && 
                     length($interface->{new_system_name})) {
                   $self->system->name($interface->{new_system_name});
                   $top->grabRelease;
                   $top->destroy;
                   $self->interface('view_state', 'task list');
                   $self->switch_systems;
                   $self->status('New system created.');
                 }
               }
              )
    ->grid(-row => 1, -column => 0, -columnspan => 2, -sticky => 'new');

  $top->deiconify; 
  $top->grab;
}


sub exit_program {
  my $self = shift;

  # save the current system?
  if (defined($self->system)) {
    my $answer = $self->interface('top')->messageBox(
                                                     -message => 
                                                     'Save Current System?',
                                                     -title => 'Save?',
                                                     -type => 'YesNo'
                                                    );
    if ($answer =~ /yes/i) {
      $self->save_system;
    }    
  } 

  $self->interface->{top}->destroy;
  exit;
}

###
# Task List screen routines
###

sub show_task_list {
  my $self = shift;
  my $frame = shift;

  my $system = $self->system;
  defined($system) || return;

  # put together a list of tasks to display
  my @tasks = ();
  foreach my $task_id ($system->tasks_keys) {
    my $task = RMA::Task->find($task_id);
    if (defined($task)) {
      push(@tasks, $task);
    } else {
      die "$task_id not a valid RMA::Task id";
    }
  }
  
  # sort by name
  if (@tasks) {
    @tasks = sort {$a->name cmp $b->name} @tasks;
  }

  # setup the display areas
  my $canvas = $frame->Scrolled('Canvas', -scrollbars => 'se', -width => $frame->width, -height => ($frame->height - 75));
  $canvas->grid(-row => 0, -column => 0, -sticky => 'nsew');
  $frame->gridRowconfigure(0, -weight => 1);
  $frame->gridColumnconfigure(0, -weight => 1);

  my $tasks_frame = $canvas->Frame();

  my $frame_window= $canvas->create('window', 0, 0, -window=>$tasks_frame, 
                                    -anchor=>'nw');

  # put the list in the frame
  $tasks_frame->Label(-text => 'Task Name', -relief => 'sunken', -border => 3,
                     -width => 30, -anchor => 'n')
    ->grid(-row => 0, -column => 0, -sticky => 'new');

  $tasks_frame->Label(-text => 'Pattern', -relief => 'sunken', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 1, -sticky => 'new');

  $tasks_frame->Label(-text => 'Period', -relief => 'sunken', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 2, -sticky => 'new');

  $tasks_frame->Label(-text => 'Exec Time', -relief => 'sunken', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 3, -sticky => 'new');

  $tasks_frame->Label(-text => 'Deadline', -relief => 'sunken', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 4, -sticky => 'new');

  $tasks_frame->Label(-text => 'Priority', -relief => 'sunken', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 5, -sticky => 'new');

  $tasks_frame->Label(-text => 'Resources', -relief => 'sunken', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 6, -sticky => 'new');


  my $y = 1;
  foreach my $task (@tasks) {
    $tasks_frame->Button(-text => $task->name, 
                                    -border => 2, -anchor => 'nw',
                                    -command => sub {
                                      GUI::TaskEditor->new(task => $task,
                                                           change_command => sub {
                                                             $self->change_view_state;
                                                           },
                                                          );
                                    }
                                   )
      ->grid(-row => $y, -column => 0, -sticky => 'nwe');

    $tasks_frame->Label(-text => $task->pattern, 
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 1, -sticky => 'nswe');

    $tasks_frame->Label(-text => $task->period, 
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 2, -sticky => 'nswe');

    $tasks_frame->Label(-text => $task->time, 
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 3, -sticky => 'nswe');

    $tasks_frame->Label(-text => $task->deadline,
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 4, -sticky => 'nswe');

    $tasks_frame->Label(-text => $task->priority,
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 5, -sticky => 'nswe');

    my @resource_names  = ();
    foreach my $resource_id ($task->resources_keys) {
      push(@resource_names, RMA::Resource->find($resource_id)->name);
    }
    my $resource_string = join(', ', @resource_names);
    length($resource_string) || ($resource_string = "None");

    $tasks_frame->Label(-text => $resource_string,
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 6, -sticky => 'nswe');

    $tasks_frame->Button(-text => "Remove",
                         -border => 2, -anchor => 'center',
                         -command => sub {
                           my $tasks = $self->system->tasks;
                           delete($tasks->{$task->id}); 
                           $task->forget;
                           $self->change_view_state;
                         }
                        )
      ->grid(-row => $y, -column => 7, -sticky => 'nswe');

    $y++;
  }

  # setup button frame
  my $button_frame = $frame->Frame(-relief => 'sunken', -border=>3, -height => 50);
  $button_frame->grid(-row => 1, -column => 0, -sticky => 'nsew');
  
  $button_frame->Button(-text => 'New Task', -width => 30, -command => 
                        sub {
                          my $task = RMA::Task->new;
                          $system->tasks($task->id, {});
                          GUI::TaskEditor->new(task => $task, change_command => 
                                               sub {
                                                 $self->change_view_state;
                                               },
                                  cancel_command => sub {
                                    my $tasks = $self->system->tasks;
                                    delete($tasks->{$task->id}); 
                                    $task->forget;
                                  }
                                    
                                              );
                        }
                       )->
    grid(-row => 0, -column => 0, -sticky => 'nsew');

  $frame->update;
  $canvas->configure(-scrollregion => [0, 0, $tasks_frame->reqwidth, $tasks_frame->reqheight]);
}

###
# Task List screen routines
###

sub show_resource_list {
  my $self = shift;
  my $frame = shift;

  my $system = $self->system;
  defined($system) || return;

  # put together a list of resources to display
  my @resources = ();
  foreach my $resource_id ($system->resources_keys) {
    my $resource = RMA::Resource->find($resource_id);
    if (defined($resource)) {
      push(@resources, $resource);
    } else {
      die "$resource_id not a valid RMA::Resource id";
    }
  }
  
  # sort by name
  if (@resources) {
    @resources = sort {$a->name cmp $b->name} @resources;
  }

  # setup the display areas
  my $canvas = $frame->Scrolled('Canvas', -scrollbars => 'se', -width => $frame->width, -height => ($frame->height - 75));
  $canvas->grid(-row => 0, -column => 0, -sticky => 'nsew');
  $frame->gridRowconfigure(0, -weight => 1);
  $frame->gridColumnconfigure(0, -weight => 1);

  my $resources_frame = $canvas->Frame();

  my $frame_window= $canvas->create('window', 0, 0, -window=>$resources_frame, 
                                    -anchor=>'nw');

  # put the list in the frame
  $resources_frame->Label(-text => 'Resource Name', -relief => 'sunken', -border => 3,
                     -width => 30, -anchor => 'n')
    ->grid(-row => 0, -column => 0, -sticky => 'new');

  $resources_frame->Label(-text => 'Policy', -relief => 'sunken', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 1, -sticky => 'new');

  my $y = 1;
  foreach my $resource (@resources) {
    $resources_frame->Button(-text => $resource->name, 
                                    -border => 2, -anchor => 'nw',
                                    -command => sub {
                                      GUI::ResourceEditor->new(resource => $resource,
                                                           change_command => sub {
                                                             $self->change_view_state;
                                                           },
                                                          );
                                    }
                                   )
      ->grid(-row => $y, -column => 0, -sticky => 'nwe');

    $resources_frame->Label(-text => $resource->policy, 
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 1, -sticky => 'nswe');

    $resources_frame->Button(-text => "Remove",
                         -border => 2, -anchor => 'center',
                         -command => sub {
                           my $resources = $self->system->resources;
                           delete($resources->{$resource->id}); 
                           $resource->forget;
                           $self->change_view_state;
                         }
                        )
      ->grid(-row => $y, -column => 2, -sticky => 'nswe');

    $y++;
  }

  # setup button frame
  my $button_frame = $frame->Frame(-relief => 'sunken', -border=>3, -height => 50);
  $button_frame->grid(-row => 1, -column => 0, -sticky => 'nsew');
  
  $button_frame->Button(-text => 'New Resource', -width => 30, -command => 
                        sub {
                          my $resource = RMA::Resource->new;
                          $system->resources($resource->id, {});
                          GUI::ResourceEditor->new(resource => $resource, change_command => 
                                               sub {
                                                 $self->change_view_state;
                                               },
                                  cancel_command => sub {
                                    my $resources = $self->system->resources;
                                    delete($resources->{$resource->id}); 
                                    $resource->forget;
                                  }
                                                   
                                                  );
                        }
                       )->
    grid(-row => 0, -column => 0, -sticky => 'nsew');

  $frame->update;
  $canvas->configure(-scrollregion => [0, 0, $resources_frame->reqwidth, $resources_frame->reqheight]);
}

###
# Resource Usage Layout
###

sub show_resource_usage_layout {
  my $self = shift;
  my $frame = shift;

  my $system = $self->system;
  defined($system) || return;

  # setup the display area - a scrolled canvas with an arbitary
  # 2000x2000 display area.
  my $canvas_frame = $frame->Frame(-width => $frame->width, -height => ($frame->height - 75))->grid(-row => 0, -column => 0, -sticky => 'nsew');

  my $canvas = $canvas_frame->Canvas(-width => $frame->width, -height => ($frame->height - 75), -scrollregion => [0,0,1024,1024]);
  my $esb = $canvas_frame->Scrollbar(-orient => 'vertical', -command => ['yview', $canvas]);
  my $ssb = $canvas_frame->Scrollbar(-orient => 'horizontal', -command => ['xview', $canvas]);
  $canvas->configure( -yscrollcommand => ['set', $esb],
                      -xscrollcommand => ['set', $ssb]
                    );

  $self->interface('canvas', $canvas);
  $canvas->grid(-row => 0, -column => 0, -sticky => 'nsew');
  $esb->grid(-row => 0, -column => 1, -sticky => 'nsw');
  $ssb->grid(-row => 1, -column => 0, -sticky => 'new');

  $frame->gridRowconfigure(0, -weight => 1);
  $frame->gridColumnconfigure(0, -weight => 1);
  $canvas_frame->gridRowconfigure(0, -weight => 1);
  $canvas_frame->gridColumnconfigure(0, -weight => 1);

  # used in bind subs for object movement
  my %object_movement;

  # used in bind subs to draw resource usage lines
  my %line_info;

  my $size_x = 200; 
  my $size_y = 100;
  my $x_offset = 50;

  # layout the tasks
  my $x_pos = 100;
  my $y_pos = 10;

  foreach my $task_id ($system->tasks_keys) {
    my $task = RMA::Task->find($task_id);
    my $task_interface = $system->tasks($task_id);
    if (!exists($task_interface->{x}) || !defined($task_interface->{x})) {
      $task_interface->{x} = $x_pos;
    } 
    if (!exists($task_interface->{'y'}) || !defined($task_interface->{'y'})) {
      $task_interface->{'y'} = $y_pos;
    } 
    if (!exists($task_interface->{'color'})) {
      $task_interface->{color} = 'white';
    }
    if (!exists($task_interface->{text_color})) {
      $task_interface->{text_color} = 'black';
    }

    # draw the task object:
    my $x = $task_interface->{x};
    my $y = $task_interface->{'y'};

    $canvas->createPolygon(($x + $x_offset), $y, 
                           ($x + $size_x + $x_offset), $y, 
                           ($x + $size_x), ($y + $size_y), 
                           $x, ($y + $size_y), 
                           -fill => $task_interface->{color}, 
                           -outline => 'black', 
                           -tags => ["TASK_${task_id}", "TASKS"]
                          );

    $canvas->createText(($x + ($size_x / 2)),($y + ($size_y / 2)), 
                        -text => $task->name,
                        -fill => $task_interface->{text_color},
                        -tags => ["TASK_${task_id}", "TASKS"],
                        -width => ($size_x - ($size_x / 4)),
                        -anchor => 'center',
                       );

    # binds to move tasks around
    $canvas->bind("TASK_${task_id}", "<Button-3>", 
                  [ sub {
                      my $self = shift;                    
                      my ($x, $y) = @_;
                      $object_movement{tag} = "TASK_" . $task_id;
                      $object_movement{task_id} = $task_id;
                      $object_movement{x} = $x;
                      $object_movement{'y'} = $y;
                      $object_movement{x_store_loc} = \$task_interface->{x};
                      $object_movement{y_store_loc} = \$task_interface->{'y'};
                      $canvas->raise($object_movement{tag});
                    },
                    Tk::Ev('x'), Tk::Ev('y')
                  ]
                 );

    # binds to draw the lines
    $canvas->bind("TASK_${task_id}", "<Button-1>", 
                  [ sub {
                      my $self = shift;                    
                      my ($x, $y) = @_;
                      $line_info{from_tag} = "TASK_" . $task_id;
                      $line_info{task_id} = $task_id;
                      $line_info{x} = $canvas->canvasx($x);
                      $line_info{'y'} = $canvas->canvasy($y);
                      delete($line_info{line_id});
                    },
                    Tk::Ev('x'), Tk::Ev('y')
                  ]
                 );

    # bind to open an editor
    $canvas->bind("TASK_${task_id}", "<Double-Button-1>", 
                  sub {
                    GUI::TaskEditor->new(task => $task,
                                         change_command => sub {
                                           $self->change_view_state;
                                         },
                                        );
                  }
                 );
                  

    $y_pos += ($size_y + ($size_y / 2));
  }

  # layout the resources
  $y_pos = 10;
  $x_pos = 350;
  foreach my $resource_id ($system->resources_keys) {
    my $resource = RMA::Resource->find($resource_id);

    my $resource_interface = $system->resources($resource_id);
    if (!exists($resource_interface->{x})) {
      $resource_interface->{x} = $x_pos;
    } 
    if (!exists($resource_interface->{'y'})) {
      $resource_interface->{'y'} = $y_pos;
    } 
    if (!exists($resource_interface->{color})) {
      $resource_interface->{color} = 'black';
    }
    if (!exists($resource_interface->{text_color})) {
      $resource_interface->{text_color} = 'white';
    }

    # draw the resource object:
    my $x = $resource_interface->{x};
    my $y = $resource_interface->{'y'};

    $canvas->createPolygon(($x + $x_offset), $y, 
                           ($x + $size_x + $x_offset), $y, 
                           ($x + $size_x), ($y + $size_y), 
                           $x, ($y + $size_y), 
                           -fill => $resource_interface->{color}, 
                           -outline => 'black', 
                           -tags => "RESOURCE_${resource_id}"
                          );

    $canvas->createText(($x + ($size_x / 2)),($y + ($size_y / 2)), 
                        -text => $resource->name,
                        -fill => $resource_interface->{text_color},
                        -tags => ["RESOURCE_${resource_id}", "RESOURCES"],
                        -width => ($size_x - ($size_x / 4)),
                        -anchor => 'center',
                       );
    $canvas->bind("RESOURCE_${resource_id}", "<Button-3>", 
                  [ sub {
                      my $self = shift;                    
                      my ($x, $y) = @_;
                      $object_movement{tag} = "RESOURCE_" . $resource_id;
                      $object_movement{resource_id} = $resource_id;
                      $object_movement{x} = $x;
                      $object_movement{'y'} = $y;
                      $object_movement{x_store_loc} = \$resource_interface->{x};
                      $object_movement{y_store_loc} = \$resource_interface->{'y'};
                      $canvas->raise($object_movement{tag});
                    },
                    Tk::Ev('x'), Tk::Ev('y')
                  ]
                 );

    $canvas->bind("RESOURCE_${resource_id}", "<Double-Button-1>", 
                  sub {
                    GUI::ResourceEditor->new(resource => $resource,
                                         change_command => sub {
                                           $self->change_view_state;
                                         },
                                        );
                  }
                 );
                        
    $y_pos += ($size_y + ($size_y / 2));
  }

  # global canvas bindings to move the objects around.

  $canvas->CanvasBind("<ButtonRelease-3>", 
                      sub {
                        if (exists($object_movement{x_store_loc}) && 
                            exists($object_movement{y_store_loc})) {
                          (${$object_movement{x_store_loc}}, ${$object_movement{y_store_loc}}, undef, undef) = $canvas->coords($object_movement{tag});
                        }
                        delete($object_movement{tag});
                        delete($object_movement{task_id});
                        delete($object_movement{resource_id});
                        $self->drawResourceUsageLines;
                      }
                     );

  # end of the line
  $canvas->CanvasBind("<ButtonRelease-1>", 
                      [
                       sub {
                         my $canvas = shift;
                         my $self = shift;
                         my ($x, $y) = @_;
                         $x = $canvas->canvasx($x);
                         $y = $canvas->canvasy($y);
                         if (!exists($line_info{from_tag})) {
                           return;
                         }
                         
                         # look for a resource to latch onto
                         my $resource_id;

                         # compile a list of candidate tags
                         my @tags;
                         my $start_tag = $canvas->find('closest', $x, $y);
                         push(@tags, $start_tag);
                         my $first_tag = $start_tag;
                         my $next_tag;
                         while(($next_tag = $canvas->find('closest', $x, $y, 0, $start_tag))) {
                           defined($next_tag) || last;
                           ($next_tag eq $first_tag) && last;
                           push(@tags, $next_tag);
                           $start_tag = $next_tag;
                         }

                         # look for a resource in @tags
                         TAG_SEARCH: foreach my $tag (@tags) {
                           my @taglist = $canvas->gettags($tag);
                           foreach my $candidate (@taglist) {
                             if ($candidate =~ /^RESOURCE_(\d+)$/) {
                               $resource_id = $1;
                               last TAG_SEARCH;
                             }
                           }
                         }

                         if (defined($resource_id)) {
                           $self->addResourceLink($line_info{task_id}, $resource_id);
                         }
                         
                         delete($line_info{from_tag});
                         if (exists($line_info{line_id})) {
                           $canvas->delete($line_info{line_id});
                         }
                       }, $self, Tk::Ev('x'), Tk::Ev('y')
                      ]
                     );
  
  $canvas->CanvasBind("<Motion>", 
                      [ sub {
                          my $self = shift;
                          $self = shift;
                          my ($x, $y) = @_;
                          if (exists($object_movement{tag})) {
                            my $dx = ($x - $object_movement{x});
                            my $dy = ($y - $object_movement{'y'});
                            $canvas->move($object_movement{tag}, $dx, $dy);
                            $object_movement{x} = $x;
                            $object_movement{'y'} = $y;
                            if (defined($object_movement{task_id})) {
                              $self->drawResourceUsageLines($object_movement{task_id});
                            } elsif (defined($object_movement{resource_id})) {
                              $self->drawResourceUsageLines;
                            }
                            $canvas->raise($object_movement{tag});
                          } elsif (exists($line_info{from_tag})) {
                            $x = $canvas->canvasx($x);
                            $y = $canvas->canvasy($y);
                            if (exists($line_info{line_id})) {
                              $canvas->delete($line_info{line_id});
                            }
                            $line_info{line_id} = $canvas->createLine($line_info{x}, $line_info{y}, $x, $y, -arrow => 'last');
                          }
                        },
                        $self, Tk::Ev('x'), Tk::Ev('y')
                      ]
                     );

  $self->drawResourceUsageLines;

  my $button_frame = $frame->Frame(-relief => 'sunken', -border=>3, -height => 50);
  $button_frame->grid(-row => 1, -column => 0, -sticky => 'nsew');
  
  $button_frame->Button(-text => 'New Task', -width => 30, -command => 
                        sub {
                          my $task = RMA::Task->new;
                          $system->tasks($task->id, {});
                          GUI::TaskEditor->new(task => $task, change_command => 
                                               sub {
                                                 $self->change_view_state;
                                               },
                                  cancel_command => sub {
                                    my $tasks = $self->system->tasks;
                                    delete($tasks->{$task->id}); 
                                    $task->forget;
                                  }
                                                   
                                                  );
                        }
                       )->
    grid(-row => 0, -column => 0, -sticky => 'nsew');

  $button_frame->Button(-text => 'New Resource', -width => 30, -command => 
                        sub {
                          my $resource = RMA::Resource->new;
                          $system->resources($resource->id, {});
                          GUI::ResourceEditor->new(resource => $resource, change_command => 
                                               sub {
                                                 $self->change_view_state;
                                               },
                                  cancel_command => sub {
                                    my $resources = $self->system->resources;
                                    delete($resources->{$resource->id}); 
                                    $resource->forget;
                                  }
                                                   
                                                  );
                        }
                       )->
    grid(-row => 0, -column => 1, -sticky => 'nsew');

  

}

sub addResourceLink {
  my $self = shift;
  my ($task_id, $resource_id) = @_;
  my $task = RMA::Task->find($task_id);
  if (!defined($task)) {
    warn "addResourceLink called with invalid task_id $task_id";
    return;
  }

  # setup linkage
  my $resources = $task->resources;
  (!exists($resources->{$resource_id})) || return;
  $resources->{$resource_id} = {};

  # call resource usage editor
  GUI::ResourceUsageEditor->new(task => $task, resource_id => $resource_id,
                                change_command => sub {
                                  $self->drawResourceUsageLines($task_id);
                                },
                                cancel_command => sub {}
                               );

  # draw the line
  $self->drawResourceUsageLines($task_id);
}

# redraws resource usage lines for specified tasks, or all tasks if
# none specified.
sub drawResourceUsageLines {
  my $self = shift;
  my $canvas = $self->interface('canvas');
  my @task_ids = @_;
  if (!scalar(@task_ids)) {
    (@task_ids = $self->system->tasks_keys);
    $canvas->delete("RESOURCE_USAGE_LINE");
  }

  foreach my $task_id (@task_ids) {
    my $resources = RMA::Task->find($task_id)->resources;

    # get task location and start point
    my ($x1, $y1, $x2, $y2) = $canvas->bbox('TASK_' . $task_id);
    my $start_x = ($x1 + (abs($x2 - $x1) / 2));
    my $start_y = ($y1 + (abs($y2 - $y1) / 2));

    foreach my $resource_id (keys %$resources) {
      my ($rx1, $ry1, $rx2, $ry2) = $canvas->bbox('RESOURCE_' . $resource_id);
      my $end_x = ($rx1 + (abs($rx2 - $rx1) / 2));
      my $end_y = ($ry1 + (abs($ry2 - $ry1) / 2));

      $canvas->delete("LINE_TASK_${task_id}_RESOURCE_${resource_id}");

      my $id = $canvas->createLine($start_x, $start_y, $end_x, $end_y, 
                                   -tags => ["LINE_TASK_${task_id}_RESOURCE_${resource_id}", "RESOURCE_USAGE_LINE"],
                                   -width => 3
                         );
      if (exists($resources->{$resource_id}{time})) {
        my ($text_x, $text_y);
        if ($start_x <= $end_x) {
          $text_x = ($start_x + (($end_x - $start_x) / 2));
        } else {
          $text_x = ($end_x + (($start_x - $end_x) / 2));
        }
        if ($start_y <= $end_y) {
          $text_y = ($start_y + (($end_y - $start_y) / 2));
        } else {
          $text_y = ($end_y + (($start_y - $end_y) / 2));
        }
        $canvas->createPolygon(($text_x - 25), ($text_y - 10),
                               ($text_x - 25), ($text_y + 10),
                               ($text_x + 25), ($text_y + 10),
                               ($text_x + 25), ($text_y - 10),
                               -fill => 'white',
                               -outline => 'white',
                               -tags => ["LINE_TASK_${task_id}_RESOURCE_${resource_id}", "RESOURCE_USAGE_LINE"],
                              );
        $canvas->createText($text_x, $text_y,
                            -text => $resources->{$resource_id}{time},
                            -anchor => 'center',
                            -width => 50,
                            -tags => ["LINE_TASK_${task_id}_RESOURCE_${resource_id}", "RESOURCE_USAGE_LINE"],
                         );
      }

      $canvas->bind("LINE_TASK_${task_id}_RESOURCE_${resource_id}", '<Button-1>',
                    sub {
                      GUI::ResourceUsageEditor->new(task => RMA::Task->find($task_id), 
                                                    resource_id => $resource_id,
                                                    change_command => sub {},
                                                    cancel_command => sub {}
                                                   );
                    }
                   );
      $canvas->bind("LINE_TASK_${task_id}_RESOURCE_${resource_id}", '<Button-3>',
                    sub {
                      delete($resources->{$resource_id});
                      $self->drawResourceUsageLines;
                    }
                   );
      $canvas->lower("LINE_TASK_${task_id}_RESOURCE_${resource_id}");
      $canvas->lower($id);
    }
  }
}


###
# Timing Analysis
###

sub show_timing_analysis {
  my $self = shift;
  my $frame = shift;

  my $system = $self->system;
  defined($system) || return;

  my $analyzer = RMA::Analyzer->new(system => $system,
                                    error_callback => 
                                    sub {
                                      my $msg = shift;
                                      $frame->messageBox(-message => $msg,
                                                         -title => 
                                                         'Analysis Engine Error',
                                                         -type => 'OK');
                                    }
                                   );
  $analyzer->analyze;
  
  # put together a list of tasks to display
  my @tasks = ();
  foreach my $task_id ($system->tasks_keys) {
    my $task = RMA::Task->find($task_id);
    if (defined($task)) {
      push(@tasks, $task);
    } else {
      die "$task_id not a valid RMA::Task id";
    }
  }
  
  # sort by priority
  if (@tasks) {
    @tasks = sort {$a->priority <=> $b->priority} @tasks;
  }
  
  # setup the display areas
  my $canvas = $frame->Scrolled('Canvas', -scrollbars => 'se', -width => $frame->width, -height => ($frame->height - 75));
  $canvas->grid(-row => 0, -column => 0, -sticky => 'nsew');
  $frame->gridRowconfigure(0, -weight => 1);
  $frame->gridColumnconfigure(0, -weight => 1);
  
  my $tasks_frame = $canvas->Frame();

  my $frame_window= $canvas->create('window', 0, 0, -window=>$tasks_frame, 
                                    -anchor=>'nw');

  # put the list in the frame
  $tasks_frame->Label(-text => 'Task Name', -relief => 'sunken', -border => 3,
                      -width => 20, -anchor => 'n')
    ->grid(-row => 0, -column => 0, -sticky => 'new');
  
  $tasks_frame->Label(-text => 'Will Meet Deadlines', -relief => 'ridge', 
                      -border => 3,
                      -width => 20, -anchor => 'n')
    ->grid(-row => 0, -column => 1, -sticky => 'new');
  
  $tasks_frame->Label(-text => 'Worst Response Time', -relief => 'ridge', 
                      -border => 3,
                      -width => 20, -anchor => 'n')
    ->grid(-row => 0, -column => 2, -sticky => 'new');

  $tasks_frame->Label(-text => 'Blocking', -relief => 'ridge', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 3, -sticky => 'new');
  
  $tasks_frame->Label(-text => 'Priority', -relief => 'sunken', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 4, -sticky => 'new');

  $tasks_frame->Label(-text => 'Pattern', -relief => 'sunken', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 5, -sticky => 'new');
  
  $tasks_frame->Label(-text => 'Exec Time', -relief => 'sunken', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 6, -sticky => 'new');
  
  $tasks_frame->Label(-text => 'Period', -relief => 'sunken', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 7, -sticky => 'new');
  
  $tasks_frame->Label(-text => 'Deadline', -relief => 'sunken', 
                      -border => 3,
                      -width => 10, -anchor => 'n')
    ->grid(-row => 0, -column => 8, -sticky => 'new');

  my $y = 1;
  foreach my $task (@tasks) {
    $tasks_frame->Button(-text => $task->name, 
                                    -border => 2, -anchor => 'nw',
                                    -command => sub {
                                      GUI::TaskEditor->new(task => $task,
                                                           change_command => sub {
                                                             $self->change_view_state;
                                                           },
                                                          );
                                    }
                                   )
      ->grid(-row => $y, -column => 0, -sticky => 'nwe');

    my $schedulable = $analyzer->isTaskSchedulable($task->id);
    if (!defined($schedulable)) { $schedulable = '?'; }
    elsif ($schedulable == 0) { $schedulable = 'No'; }
    elsif ($schedulable == 1) { $schedulable = 'Yes'; }
    else { $schedulable = '?'; }
    $tasks_frame->Label(-text => $schedulable,
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 1, -sticky => 'nswe');

    my $response_time = $analyzer->taskResponseTime($task->id);
    if (!defined($response_time)) { $response_time = '?'; }
    $tasks_frame->Label(-text => $response_time,
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 2, -sticky => 'nswe');

    my $blocking = $analyzer->taskBlocking($task->id);
    defined($blocking) || ($blocking = '0');
    $tasks_frame->Label(-text => $blocking,
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 3, -sticky => 'nswe');

    $tasks_frame->Label(-text => $task->priority, 
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 4, -sticky => 'nswe');

    $tasks_frame->Label(-text => $task->pattern, 
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 5, -sticky => 'nswe');

    $tasks_frame->Label(-text => $task->time, 
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 6, -sticky => 'nswe');

    $tasks_frame->Label(-text => $task->period, 
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 7, -sticky => 'nswe');

    $tasks_frame->Label(-text => $task->deadline,
                        -border => 2, -anchor => 'center')
      ->grid(-row => $y, -column => 8, -sticky => 'nswe');

    $y++;
  }

  my $button_frame = $frame->Frame(-relief => 'sunken', -border=>3, -height => 50);
  $button_frame->grid(-row => 1, -column => 0, -sticky => 'nsew');
  
  $button_frame->Button(-text => 'Explain Results', -width => 30, 
                        -command => [\&explain_results_screen, $self, $analyzer],
                       )->
                         grid(-row => 0, -column => 0, -sticky => 'nsew');
  
  $frame->update;
  $canvas->configure(-scrollregion => [0, 0, $tasks_frame->reqwidth, $tasks_frame->reqheight]);
}

sub explain_results_screen {
  my $self = shift;
  my $analyzer = shift;
  my $top = MainWindow->new;
  $top->configure(-title => 'gRMA : ' . $self->system->name . ' : Explain Results');
  $top->geometry('+150+110');
  my $text = $top->Scrolled('Text', -scrollbars => 'e', 
                            -height => 24, -width => 80, 
                            -wrap => 'word',
                            -font => ['helvetica', 12],
                           );
  
  my @tasks;
  foreach my $task_id ($self->system->tasks_keys) {
    push(@tasks, RMA::Task->find($task_id));
  }
  @tasks = sort {$a->priority <=> $b->priority} @tasks;

  foreach my $task (@tasks) {
    $text->insert('end', "Task Name: " . $task->name . "\n");
    $text->insert('end', "Analysis Explanation: " . $analyzer->taskExplanation($task->id) . "\n\n");
  }
  my $close_button = $top->Button(-text => "Close", 
                                  -command => sub { $top->destroy; },
                                 );

  $text->configure(-state => 'disabled');
  $text->grid(-row => 0, -column => 0, -sticky => 'nsew');
  $close_button->grid(-row => 1, -column => 0, -sticky => 's');
  $top->gridRowconfigure(0, -weight => 1);
  $top->gridColumnconfigure(0, -weight => 1);
}


1;
__END__
=head1 NAME

GUI::MainScreen - the main screen routines for the gRMA system

=head1 SYNOPSIS

  use GUI::MainScreen;
  use Tk;

  GUI::MainScreen->new;
  MainLoop;

=head1 DESCRIPTION

These are the routines for the main screen of gRMA.  Everything that
happens in the main screen of the gRMA program is in this file!  It
also handles saving and loading of the RMA::* object data.

=head1 AUTHOR

Sam Tregar, sam@tregar.com

=cut
