package GUI::TaskEditor;

use strict;
use vars qw($VERSION @ISA);
use Tk;
use RMA::Task;

use GUI::Help;

use Class::MethodMaker
  new_with_init => 'new',
  new_hash_init => 'hash_init',
  get_set => [ 'title', 'task', 'state', 'change_command', 'cancel_command'],
  hash => [ 'interface' ],
;

$VERSION = '0.01';

sub init {
  my $self = shift;
  $self->hash_init(@_);

  # initialize managed data
  $self->init_data;

  # initialize the window
  $self->init_window;

  return $self;
}

sub init_data {
  my $self = shift;

  # if we weren't passed a task to edit, make a new one
  # reflect that in the title if it ain't set
  if (!defined($self->task)) {
    $self->task(RMA::Task->new);
    if (!defined($self->title)) {
      $self->title('Task Editor : New Task');
    }
  }

  return $self;
}

sub init_window {
  my $self = shift;

  my $interface = $self->interface;
  if (!exists($interface->{top})) {
    $interface->{top} = MainWindow->new();
  } 
  # get rid of any existing contents if passed a toplevel
  my $top = $interface->{top};
  $top->geometry('+100+100');

  # setup the title
  if (defined($self->title)) {
    $top->title($self->title);
  } else {
    $top->title('Task Editor');
  }

  my $main_frame = $top->Frame()->grid(-row => 0, -column => 0, -sticky => 'nsew');
  $interface->{main_frame} = $main_frame;
  my $button_frame = $top->Frame()->grid(-row => 1, -column => 0, -sticky => 'sew');
  $self->interface('button_frame', $button_frame);

  $top->gridRowconfigure(0, -weight => 1);
  $top->gridRowconfigure(1, -weight => 0);
  $top->gridColumnconfigure(0, -weight => 1);

  my $name_frame = $main_frame->Frame(-border => 3, -relief => 'ridge');
  my $name_label = $name_frame->Label(-text => 'Task Name');
  my $name_entry = $name_frame->Entry();
  my $name_help = $name_frame->Button(-text => '?', 
                                      command => 
                                      [\&showHelp, 'TaskEditor#name']
                                     );
  $name_entry->insert(0, $self->task()->name());

  $name_label->grid(-row => 0, -column => 0, -sticky => 'nw');
  $name_entry->grid(-row => 1, -column => 0, -sticky => 'sew');
  $name_help->grid(-row => 1, -column => 1, -sticky => 'se');

  $name_frame->gridRowconfigure(0, -weight => 1);
  $name_frame->gridRowconfigure(1, -weight => 0);
  $name_frame->gridColumnconfigure(0, -weight => 1);
  $name_frame->gridColumnconfigure(1, -weight => 0);

  $self->interface('saved_pattern', $self->task->pattern);

  my $pattern_frame = $main_frame->Frame(-border => 3, -relief => 'ridge');
  my $pattern_label = $pattern_frame->Label(-text => 'Execution Pattern');
  my $pattern_list = $pattern_frame->Listbox(-width=>20, -height=> 5, -exportselection => 0);
  $pattern_list->insert('end', "Periodic", "Bursty", "Bounded", "Irregular", "Unbounded");
  if (!defined($self->task->pattern)) {
    $pattern_list->selectionSet(0);
    $self->task->pattern('Periodic');
  } elsif ($self->task->pattern eq 'Periodic')  {
    $pattern_list->selectionSet(0);
  } elsif ($self->task->pattern eq 'Bursty') {
    $pattern_list->selectionSet(1);
  } elsif ($self->task->pattern eq 'Bounded') {
    $pattern_list->selectionSet(2);
  } elsif ($self->task->pattern eq 'Irregular') {
    $pattern_list->selectionSet(3);
  } elsif ($self->task->pattern eq 'Unbounded') {
    $pattern_list->selectionSet(4);
  } else {
    $pattern_list->selectionSet(0);
    $self->task->pattern('Periodic');
  }

  $pattern_list->bind('<Button-1>', 
                      sub {
                        my $pattern = $pattern_list->get($pattern_list->curselection);
                        if (!$pattern) {
                          $pattern_list->selectionSet(0);
                          $pattern = 'Periodic';
                        } 
                        $self->task->pattern($pattern);
                      }
                     );
                        
  my $pattern_help = $pattern_frame->Button(-text => '?', -command => [\&showHelp, 'TaskEditor#pattern']);

  $pattern_label->grid(-row => 0, -column => 0, -sticky => 'nw');
  $pattern_list->grid(-row => 1, -column => 0, -sticky => 'sew');
  $pattern_help->grid(-row => 1, -column => 1, -sticky => 'se');

  $pattern_frame->gridRowconfigure(0, -weight => 1);
  $pattern_frame->gridRowconfigure(1, -weight => 0);
  $pattern_frame->gridColumnconfigure(0, -weight => 1);
  $pattern_frame->gridColumnconfigure(1, -weight => 0);

  my $period_frame = $main_frame->Frame(-border => 3, -relief => 'ridge');
  my $period_label = $period_frame->Label(-text => 'Period');
  my $period_entry = $period_frame->Entry(-width => 4);
  my $period_help = $period_frame->Button(-text => '?', 
                                      command => 
                                      [\&showHelp, 'TaskEditor#period']
                                     );
  $period_entry->insert(0, $self->task()->period());

  $period_label->grid(-row => 0, -column => 0, -sticky => 'nw');
  $period_entry->grid(-row => 1, -column => 0, -sticky => 'sew');
  $period_help->grid(-row => 1, -column => 1, -sticky => 'se');

  $period_frame->gridRowconfigure(0, -weight => 1);
  $period_frame->gridRowconfigure(1, -weight => 0);
  $period_frame->gridColumnconfigure(0, -weight => 1);
  $period_frame->gridColumnconfigure(1, -weight => 0);

  my $time_frame = $main_frame->Frame(-border => 3, -relief => 'ridge');
  my $time_label = $time_frame->Label(-text => 'Execution Time');
  my $time_entry = $time_frame->Entry(-width => 4);
  my $time_help = $time_frame->Button(-text => '?', 
                                      command => 
                                      [\&showHelp, 'TaskEditor#time']
                                     );
  $time_entry->insert(0, $self->task()->time());

  $time_label->grid(-row => 0, -column => 0, -sticky => 'nw');
  $time_entry->grid(-row => 1, -column => 0, -sticky => 'sew');
  $time_help->grid(-row => 1, -column => 1, -sticky => 'se');

  $time_frame->gridRowconfigure(0, -weight => 1);
  $time_frame->gridRowconfigure(1, -weight => 0);
  $time_frame->gridColumnconfigure(0, -weight => 1);
  $time_frame->gridColumnconfigure(1, -weight => 0);

  my $deadline_frame = $main_frame->Frame(-border => 3, -relief => 'ridge');
  my $deadline_label = $deadline_frame->Label(-text => 'Deadline');
  my $deadline_entry = $deadline_frame->Entry(-width => 4);
  my $deadline_help = $deadline_frame->Button(-text => '?', 
                                      command => 
                                      [\&showHelp, 'TaskEditor#deadline']
                                     );
  $deadline_entry->insert(0, $self->task()->deadline());

  $deadline_label->grid(-row => 0, -column => 0, -sticky => 'nw');
  $deadline_entry->grid(-row => 1, -column => 0, -sticky => 'sew');
  $deadline_help->grid(-row => 1, -column => 1, -sticky => 'se');

  $deadline_frame->gridRowconfigure(0, -weight => 1);
  $deadline_frame->gridRowconfigure(1, -weight => 0);
  $deadline_frame->gridColumnconfigure(0, -weight => 1);
  $deadline_frame->gridColumnconfigure(1, -weight => 0);

  my $priority_frame = $main_frame->Frame(-border => 3, -relief => 'ridge');
  my $priority_label = $priority_frame->Label(-text => 'Priority');
  my $priority_entry = $priority_frame->Entry(-width => 4);
  my $priority_help = $priority_frame->Button(-text => '?', 
                                      command => 
                                      [\&showHelp, 'TaskEditor#priority']
                                     );
  $priority_entry->insert(0, $self->task()->priority());

  $priority_label->grid(-row => 0, -column => 0, -sticky => 'nw');
  $priority_entry->grid(-row => 1, -column => 0, -sticky => 'sew');
  $priority_help->grid(-row => 1, -column => 1, -sticky => 'se');

  $priority_frame->gridRowconfigure(0, -weight => 1);
  $priority_frame->gridRowconfigure(1, -weight => 0);
  $priority_frame->gridColumnconfigure(0, -weight => 1);
  $priority_frame->gridColumnconfigure(1, -weight => 0);



  $name_frame->grid(-row => 0, -column => 0 , -columnspan => 2, -sticky => 'nsew', -padx => 3, -pady => 5);
  $pattern_frame->grid(-row => 1, -rowspan => 2, -column => 0 , -sticky => 'nsew', -padx => 3, -pady => 5);
  $period_frame->grid(-row => 1, -column => 1 , -sticky => 'nsew', -padx => 3, -pady => 5);
  $time_frame->grid(-row => 2, -column => 1 , -sticky => 'nsew', -padx => 3, -pady => 5);
  $deadline_frame->grid(-row => 3, -column => 0 , -sticky => 'nsew', -padx => 3, -pady => 5);
  $priority_frame->grid(-row => 3, -column => 1 , -sticky => 'nsew', -padx => 3, -pady => 5);


  $main_frame->gridColumnconfigure(0, -weight => 1);
  $main_frame->gridColumnconfigure(1, -weight => 1);
  $main_frame->gridRowconfigure(0, -weight => 1);
  $main_frame->gridRowconfigure(1, -weight => 1);
  $main_frame->gridRowconfigure(2, -weight => 1);
  $main_frame->gridRowconfigure(3, -weight => 1);
    
  $button_frame->Button(-text => 'Ok', -command => sub {
                          $self->task->name($name_entry->get);
                          $self->task->time($time_entry->get);
                          $self->task->period($period_entry->get);
                          $self->task->priority($priority_entry->get);
                          $self->task->deadline($deadline_entry->get);
                          $top->destroy;
                          my $callback = $self->change_command;
                          defined($callback) && (&$callback ());
                        }
                       )->grid(-row => 0, -column => 0, -sticky => 'nsew');

  $button_frame->Button(-text => 'Cancel', -command => sub {
                          $self->task->pattern($self->interface('saved_pattern'));
                          $top->destroy;
                          my $callback = $self->cancel_command;
                          defined($callback) && (&$callback ());
                        }
                       )->grid(-row => 0, -column => 1, -sticky => 'nsew');

  $button_frame->gridRowconfigure(0, -weight => 1);
  $button_frame->gridColumnconfigure(0, -weight => 1);
  $button_frame->gridColumnconfigure(1, -weight => 1);

  $top->update;
  return $self;
}

1;
__END__
# Below is the stub of documentation for your module. You better edit it!

=head1 NAME

GUI::TaskEditor - Perl extension for blah blah blah

=head1 SYNOPSIS

  use GUI::TaskEditor;
  blah blah blah

=head1 DESCRIPTION

Stub documentation for GUI::TaskEditor was created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head1 AUTHOR

A. U. Thor, a.u.thor@a.galaxy.far.far.away

=head1 SEE ALSO

perl(1).

=cut
