package RMA::Analyzer;

=head1 NAME

RMA::Analyzer - Perl module implementing the RMA analysis engine.

=head1 SYNOPSIS

  use RMA::Analyzer;
 
  # fill in an RMA::System with RMA::Tasks and RMA::Resources..
  my $analyzer = RMA::Analyzer->new(system => $system);

  # run the engine
  my $result = $analyzer->analyze();

=head1 DESCRIPTION

=cut

use Class::MethodMaker
  new_with_init => 'new',
  new_hash_init => '_hash_init',
  key_attrib => 'id',
  get_set => [ 'name', 'system', 'error_callback' ],
  hash => 'results'
;

use Carp;
use POSIX; 

use strict;
use vars qw($VERSION @ISA $DEBUG);

$VERSION = '0.01';

# The debug level - determines how much debugging information is
# printed to STDOUT.  1 prints the high level results of the analysis,
# 2 prints copious intermediate results.
$DEBUG = 0;

use RMA::Object;
@ISA = qw(RMA::Object);

# Looks at the system and decides if the system can be analyzed by
# this module.  If you add more analysis functionality, you might
# need to change this to let more varied systems through.
sub judge {
  my $self = shift;
  my $system = $self->system;

  my (%saw_priority);
  foreach my $task_id ($system->tasks_keys) {
    my $task = RMA::Task->find($task_id);

    # reject systems with resources using unsupported policies
    foreach my $resource_id ($task->resources_keys) {
      my $resource = RMA::Resource->find($resource_id);
      if (
          !($resource->policy =~ /IMP/) &&
          !($resource->policy =~ /PIP/) &&
          !($resource->policy =~ /Highest Locker/) &&
          !($resource->policy =~ /PCP/)
         ) 
        {
        $self->error("Sorry, the analyzer doesn't handle resources using the " . $resource->policy . " policy yet!  If you're a programmer, please consider contributing code to assess this lack.  If you're just a normal user, please check the next release.");
        return 0;
      }
    }

    # make sure we have distinct priorities
    if (exists($saw_priority{$task->priority})) {
      $self->error("The analyzer requires that each task be assigned a unique priority.  Please go to the Task View and resolve the conflict.");
      return 0;
    }
    $saw_priority{$task->priority} = 0;

    # only accept tasks with simple periodic translations...
    if (($task->pattern ne "Periodic") &&
        ($task->pattern ne "Bounded") &&
        ($task->pattern ne "Bursty"))
      {
        $self->error("This analyzer only handles periodic, bounded aperiodic, and bursty aperiodc tasks.  If you are a programmer you may wish to contribute to a more full-featured analyzer.  If you are a user that needs the ability to analyze a wider range of tasks, please check back for the next release.");
        return 0;      
    }
  }

  

  return 1;
}

# Runs the analyzer on a given System.  Returns 1 on success, undef on
# failure.  This must be called before any of the schedulibility
# information retrieval can be done.
sub analyze {
  my $self = shift;

  # Answer the question, "can the analyzer work with this system?"
  # Someday, if all the possible combinations of task and resource
  # settings can be addressed by this module, this call can be
  # removed.
  if (!$self->judge()) {
    return undef;
  }
  
  # check the effective utilization for the system.  If it's greater
  # than 1, then we stop - noting in the results hash that the system
  # is unschedulable.  If it's less than one, we proceed to use the
  # more precise tests to determine schedulability.  We do this test
  # first because effective utilzation of > 1 can cause an infinite
  # loop in the later tests.
  my $effective_utilization = $self->effective_utilization;
  defined($effective_utilization) || (return undef);
  if ($effective_utilization >= 1.0) {
    my $results = $self->results;
    foreach my $task_id ($self->system->tasks_keys) {
      $results->{schedulable}{$task_id} = 0;
      $results->{explanation}{$task_id} = "Task was found unschedulable with a simple effective utilization test for the entire system - irrespective of scheduling constraints, this system requires $effective_utilization times the available processing time.";
    }
    $self->results('system_schedulable', 0);
    $self->results('effective_utilization', $effective_utilization);
    return 0;
  }

  # for each task in the system, ordered by priority, we compute a
  # worst case arrival time.
  if (!$self->compute_worst_case_task_response_times()) {
    return undef;
  }

}

# computes the simple effective utilization value for a system of
# tasks.
sub effective_utilization {
  my $self = shift;
  my @tasks = $self->task_list_by_priority();

  my $U_total = 0;
  my $n = scalar(@tasks) . '.0';
  my $U_bound = ($n * ((2 ** (1 / $n)) - 1)); # not used for much but debugging

  # computer $U_total 

  for (my $x = 0; $x < $n; $x++) {
    my ($time, $period);
    if (($tasks[$x]->pattern eq 'Periodic') || ($tasks[$x]->pattern eq 'Bounded')) {
      $time = $tasks[$x]->time;
      $period = $tasks[$x]->period;
    } elsif ($tasks[$x]->pattern eq 'Bursty') {
      my ($int, $count) = split(',', $tasks[$x]->period());
      if (!defined($count) || !defined($int)) {
        $self->error("Period for task " . $tasks[$x]->name . " has a bad format - the format for a bursty task's period is interval,max_runs.  Please see the documentation for more information");
        return 0;
      }  

      $time = $tasks[$x]->time * $count;
      $period = $int;
    }
  
    $U_total += ($time / $period);
  }

  if ($DEBUG >= 1) { 
    print "Effective utilzation: $U_total\n";
    print "UB($n) = $U_bound\n";
  }

  # return effective utilization
  return $U_total;
}

# stores numeric results in $self->results('response_time') and simple
# yes/no info in $self->results('schedulable'), hashed by task_id
sub compute_worst_case_task_response_times {
  my $self = shift;
  my $results = $self->results;
  $results->{response_time} = {};
  $results->{schedulable} = {};

  my @tasks = $self->task_list_by_priority();

  # compute the blocking terms for @tasks:

  # get a list of resources used by at least one task - resource_id
  # hashes to a ref to an array of tasks
  # also initialize %blocking
  my %blocking;
  my %resources;
  my %extra_time;
  foreach my $task (@tasks) {
    foreach my $resource_id ($task->resources_keys) {
      if (!exists($resources{$resource_id})) {
        $resources{$resource_id} = [ $task ];
      } else {
        push(@{$resources{$resource_id}}, $task);
      }
    }
    $blocking{$task->id} = 0;
  }
  
  # take care of resources being used by only one task
  # just add the time used to the execution time
  foreach my $resource_id (keys %resources) {
    if (scalar(@{$resources{$resource_id}}) == 1) {
      my $task = $resources{$resource_id}->[0];
      delete($resources{$resource_id});
    }
  }
  
  # find blocking for resources using the Interupt-Masking protocol
  foreach my $resource_id (keys %resources){ 
    my $resource = RMA::Resource->find($resource_id);
    next unless ($resource->policy =~ /IMP/);

    # compute the blocking terms, bottom up for IMP - hitting all,
    # regardless of resource usage.
    my $b = 0;
    for (my $x = $#tasks; $x >= 0; $x--) {
      my $task = $tasks[$x];
      $blocking{$task->id} += $b;
      if (exists($task->resources->{$resource_id}) && 
          ($task->resources->{$resource_id}{time} > $b)) {
        $b = $task->resources->{$resource_id}{time};
      }
    }
    delete($resources{$resource_id});
  }

  # find blocking for resources using Priority Inheritence 
  foreach my $resource_id (keys %resources){ 
    my $resource = RMA::Resource->find($resource_id);
    next unless ($resource->policy =~ /PIP/);
    
    # find the priority ceiling of this resource
    my $priority_ceiling = -1;
    foreach my $task (@tasks) {
      if (exists($task->resources->{$resource_id})) {
        if ($task->priority > $priority_ceiling) {
          $priority_ceiling = $task->priority;
        }
      }
    }

    # compute the blocking terms, bottom up for PIP - hitting all but
    # the top priority task regardless of resource usage.
    my $b = 0;
    for (my $x = $#tasks; $x > 0; $x--) {
      my $task = $tasks[$x];
      last unless ($task->priority <= $priority_ceiling);
      $blocking{$task->id} += $b;
      if (exists($task->resources->{$resource_id}) && 
          ($task->resources->{$resource_id}{time} > $b)) {
        $b = $task->resources->{$resource_id}{time};
      }
    }
    delete($resources{$resource_id});
  }
  
  # find blocking for resources using Highlest Locker and PCP...
  # Apparently the same analytically.
  my %hl_pcp_blocking;
  foreach my $resource_id (keys %resources){ 
    my $resource = RMA::Resource->find($resource_id);
    next unless (($resource->policy =~ /Highest Locker/) || ($resource->policy =~ /PCP/));
    
    # find the priority ceiling of this resource
    my $priority_ceiling = -1;
    foreach my $task (@tasks) {
      if (exists($task->resources->{$resource_id})) {
        if ($task->priority > $priority_ceiling) {
          $priority_ceiling = $task->priority;
        }
      }
    }


    # compute the blocking terms, bottom up for HL and PCP - hitting
    # all but the top priority task regardless of resource usage.
    my $b = 0;
    for (my $x = $#tasks; $x > 0; $x--) {
      my $task = $tasks[$x];
      if (!exists($hl_pcp_blocking{$task->id})) {
        $hl_pcp_blocking{$task->id} = [ 0 ];
      }
      last unless ($task->priority <= $priority_ceiling);
      push(@{$hl_pcp_blocking{$task->id}}, $b);
      if (exists($task->resources->{$resource_id}) && 
          ($task->resources->{$resource_id}{time} > $b)) {
        $b = $task->resources->{$resource_id}{time};
      }
    }
    delete($resources{$resource_id});
  }

  # resolve the hl_pcp__blocking result hash to use on the max value
  foreach my $task (@tasks) {
    next unless exists($hl_pcp_blocking{$task->id});
    my $hl_pcp_blocking_value = 0;
    foreach my $value (@{$hl_pcp_blocking{$task->id}}) {
      if ($value > $hl_pcp_blocking_value) {
        $hl_pcp_blocking_value = $value;
      }
    }
    $blocking{$task->id} += $hl_pcp_blocking_value;
  }


  # stow away blocking values for retrevial:
  foreach my $task (@tasks) {
    $results->{blocking}{$task->id} = $blocking{$task->id};
  }

  # compute a response time for each task, descending by priority
  # keeping track of which tasks are above the current one as we go.
  my @tasks_above = ();
  my @tasks_above_time = ();
  my @tasks_above_period = ();
  foreach my $task (@tasks) {
    my ($time, $period);
    if (($task->pattern eq 'Periodic') || ($task->pattern eq 'Bounded')) {
      $time = $task->time;
      $period = $task->period;
    } elsif ($task->pattern eq 'Bursty') {
      my ($int, $count) = split(',', $task->period());
      if (!defined($count) || !defined($int)) {
        $self->error("Period for task " . $task->name . " has a bad format - the format for a bursty task's period is interval,max_runs.  Please see the documentation for more information");
        return 0;
      }  
      $time = $task->time * $count;
      $period = $int;
    }

    my $b = $blocking{$task->id};

    # compute the first approximation
    my $a0 = 0;
    $a0 += $b + $time;
    foreach my $above_time (@tasks_above_time) {
      $a0 += $above_time;
    }

    # store for itteration
    my $a_last = $a0;

    if ($DEBUG >= 2) {
      print $task->name, " a0: $a0\n";
    }

    # initialize a counter to represent to number of times this task
    # is run in the busy period.
    my $k = 1;

    # array to store response times for each run
    my @response_time;

    # find the worst-case response time for this task
    # iterating through each run in the busy period
    

    # don't let the calcualte_responses loop run too many times..
    my $calc_loop_limit = 1001;

  CALCULATE_RESPONSES: while (1) {
      
      # make sure we don't enter an infinite sequence... It's happened
      # a few times during testing of the IMP resource policy.
      if (!$calc_loop_limit) {
        $results->{explanation}{$task->id} = "The response time test looped more than 1000 times trying to determind the response time of this task.  This is an observed bug with no currently known solution - please help me fix it if you can, but don't send me a bug report about it.";
        return 0;
      } else {
        $calc_loop_limit--;
      }
        
      # iterate until the algorithm converges on a final value for a
      # for this run
      
    FIND_A: while(1) {
        my $a_next = 0;
        $a_next += $b;
        $a_next += ($k * $time);
        for (my $x = 0; $x <= $#tasks_above; $x++) {
          my $p = $tasks_above_period[$x];
          my $t = $tasks_above_time[$x];
          $a_next += (POSIX::ceil(($a_last / $p)) * $t);
        }
        if ($DEBUG >= 2) {
          print $task->name, " a_next: $a_next\n";
        }
        
        # have we converged?
        last FIND_A if ($a_last == $a_next);
        # nope..  store and repeat.
        $a_last = $a_next;
      }
      
      # compute the response time for this run
      $response_time[$k] = $a_last - ($period * ($k - 1));
      if ($DEBUG >= 2) {
        print "Response time $k : $response_time[$k]\n";
      }

      # are we done with the busy period?
      if ($response_time[$k] > $period) {
        # nope.
        $k++;
        $a_last = $a_last + $time;
        next CALCULATE_RESPONSES;
      } else {
        last CALCULATE_RESPONSES;
      }
    }

    # all done, get worst case
    my $worst_case_response = 0;
    for (my $x = 1; $x <= $k; $x++) {
      if ($response_time[$x] > $worst_case_response) {
        $worst_case_response = $response_time[$x];
      }
    }
    
    if ($DEBUG >= 1) {
      print $task->name, " : worst case response time : $worst_case_response\n";
    }

    # we got!
    $results->{response_time}{$task->id} = $worst_case_response;

    # support use of dealine == 0 to mean at the end of the period
    my $deadline = $task->deadline;
    if ($deadline == 0) { $deadline = $period; }
    if ($worst_case_response <= $deadline) {
      if ($DEBUG >= 1) {
        print $task->name, " : schedulable.\n";
      }
      $results->{schedulable}{$task->id} = 1;
      $results->{explanation}{$task->id} = "Task was found schedulable using the response time test for periodic tasks with arbitrary deadlines and blocking.";
    } else {
      if ($DEBUG >= 1) {
        print $task->name, " : not schedulable.\n";
      }
      $results->{schedulable}{$task->id} = 0;
      $results->{explanation}{$task->id} = "Task was found to be unschedulable using the response time test for periodic tasks with arbitrary deadlines and blocking.";
    }

    push(@tasks_above, $task);
    push(@tasks_above_time, $time);
    push(@tasks_above_period, $period);

  }

  return 1;
}

# returns a list of RMA::Task objects sorted by priority
sub task_list_by_priority {
  my $self = shift;
  my $system = $self->system;

  my @tasks = ();
  foreach my $task_id ($system->tasks_keys) {
    push(@tasks, RMA::Task->find($task_id));
  }
  @tasks = sort {$a->priority() <=> $b->priority();} @tasks;

  return @tasks;
}

###
# Information retrieval methods
###

sub isTaskSchedulable {
  my $self = shift;
  my $task_id = shift;
  my $system = $self->system;
  my $results = $self->results;
  if ((exists $results->{schedulable}) && 
      (exists $results->{schedulable}{$task_id})) {
    return $results->{schedulable}{$task_id};
  } else {
    return undef;
  }
}

sub taskResponseTime {
  my $self = shift;
  my $task_id = shift;
  my $system = $self->system;
  my $results = $self->results;
  if ((exists $results->{response_time}) && 
      (exists $results->{response_time}{$task_id})) {
    return $results->{response_time}{$task_id};
  } else {
    return undef;
  }
}

sub taskExplanation {
  my $self = shift;
  my $task_id = shift;
  my $system = $self->system;
  my $results = $self->results;
  if ((exists $results->{explanation}) && 
      (exists $results->{explanation}{$task_id})) {
    return $results->{explanation}{$task_id};
  } else {
    return '';
  }
}

sub taskBlocking {
  my $self = shift;
  my $task_id = shift;
  my $system = $self->system;
  my $results = $self->results;

  if ((exists $results->{blocking}) && 
      (exists $results->{blocking}{$task_id})) {
    return $results->{blocking}{$task_id};
  } else {
    return undef;
  }
}

###
# Utility Functions
###

# handles setting up id if necessary and calling _hash_init.  also
# checks for system param and sets up a default error_callback
sub init {
  my $self = shift;
  @_ = $self->SUPER::init(@_);
  $self->_hash_init(@_);

  if (!defined($self->error_callback)) {
    $self->error_callback(sub { die @_; });
  }
  if (!defined($self->system)) {
    die "Called RMA::System->new() without a system parameter.";
  }
}

# calls $self->error_callback with specified error message
sub error {
  my ($self, $msg) = @_;
  my $error_callback = $self->error_callback;
  &$error_callback ($msg);
}



1;

=head1 AUTHOR

Sam Tregar, sam@tregar.com

=cut
