package RMA::Object;
=head1 NAME

RMA::Object - Perl extension to provide a common base for the
RMA::* object classes

=head1 SYNOPSIS

  use RMA::Object;
  use Class::MethodMaker
      key_attrib => 'id';

  @ISA = (RMA::Object);



=head1 DESCRIPTION

This module provides a few object methods for all the RMA::* objects.

Things you MUST do to inherit successfully from this class:

=over 4

=item *
use Class::MethodMaker

=over 8

=item *
declare just one ket_attrib called 'id'.

=item *
declare a new_hash_init called '_hash_init'

=item * make an init sub that looks like:

   sub init {
       my $self = shift;
       @_ = $self->SUPER::init(@_);

       # your init stuff here

       $self->_hash_init(@_);
   }

=back

=back

Currently this module provides init - a function you should call
from YOUR init routine to setup the id attribute, dump, dumpAll, and
load.  These last few provide rudimentary object persistence.

See the source for RMA::Task.pm for a good authoritative usage.

=cut

use strict;
use vars qw($VERSION);

$VERSION = '0.01';

use Carp;
use Data::Dumper;

# handles setting up id if necessary and calling _hash_init
sub init {
  my $self = shift;
  
  # find out if we need to set id...
  my $id_already_set = 0;
  if (@_) {
    my %initItems = @_;
    if (exists($initItems{id})) {
      $id_already_set = 1;
      # check that this id is free for use
      if (defined($self->find_id($initItems{id}))) {
        croak "Attempt to create an RMA object with id $initItems{id} - already in use!";
      }
    }
  }

  # if we need to set id...
  if (!$id_already_set) {
    # look for a free id...
    my $id = 1;
    while(defined($self->find_id($id))) {
      $id++;
    }
    push(@_, 'id', $id);
  }
  return @_;
}

# simple shorthand for find_id
sub find {
  my $self = shift;
  $self->find_id(@_);
}

sub forget {
  my $self = shift;
  my $list = $self->find;
  delete($list->{$self->id});
  return $self;
}

sub dumpAll {
  my $self = shift;
  return Data::Dumper->Dumpxs([$self->find], ['dumpRef']);
}

sub dump {
  my $self = shift;
  return Data::Dumper->Dumpxs([{$self->id() => $self}], ['dumpRef']);
}

sub load {
  my $self = shift;
  my $list = $self->find;
  foreach my $dump (@_) {
    my $dumpRef;
    eval $dump;
    my ($key, $val);
    while (($key, $val) = each %{$dumpRef}) { $list->{$key} = $val; }
  }
}


=head1 AUTHOR

Sam Tregar, sam@tregar.com

=cut

1;
