package RMA::Task;

=head1 NAME

RMA::Task - Perl class representing a task in the RMA module

=head1 SYNOPSIS

  use RMA::Task;
  my $task = RMA::Task->new(period => 10, time => 100, name => "Task One");
  my $task_id = RMA::Task->task_id;

  my $found_task = RMA::Task->find($task_id);

=head1 DESCRIPTION

This class represents a single task (or "process") in the RMA module
system.  A task in terms of Rate Monotonic Analysis is periodic and
bounded in terms of worst-case timing.  

The period of a task in milliseconds is stored in the "period"
attribute and the worst-case computing time is stored in the "time"
attribute.  Additionaly RMA::Task supports a "priority" attribute.
This attribute is usually set by some scheduling policy, enforced by
other RMA objects.  It is located in the Task object for convenience.
See below for more detauls concerning these attributes.

An RMA::Task may be associated with any number of RMA::Resource
objects through the resources hash of hashes.  See below for more
details.

=cut

use Class::MethodMaker
  new_with_init => 'new',
  new_hash_init => '_hash_init',
  key_attrib => 'id',
  get_set => [ 'name', 'pattern', 'period', 'deadline', 'time', 'priority', 
             ],
  hash => 'resources',
;

use Carp;

use strict;
use vars qw($VERSION @ISA);

$VERSION = '0.01';

use RMA::Object;
@ISA = qw(RMA::Object);

# handles setting up id if necessary and calling _hash_init

sub init {
  my $self = shift;
  @_ = $self->SUPER::init(@_);
  $self->_hash_init(@_);
  defined($self->name) || $self->pattern('');
  defined($self->pattern) || $self->pattern('Periodic');
  defined($self->period) || $self->period(0);
  defined($self->deadline) || $self->deadline(0);
  defined($self->time) || $self->time(0);
  defined($self->priority) || $self->priority(0);
}

1;
__END__

=head1 AUTHOR

Sam Tregar, sam@tregar.com

=cut
